import logging
from datetime import datetime
from datetime import timezone
from typing import List
from typing import Optional
from typing import Union

from asyncio_mqtt import Client
from senor_octopus.types import Stream

_logger = logging.getLogger(__name__)


async def mqtt(
    topics: Union[str, List[str]],
    host: str = "localhost",
    port: str = "1883",
    username: Optional[str] = None,
    password: Optional[str] = None,
    prefix: str = "hub.mqtt",
) -> Stream:
    _logger.info("Subscribing to MQTT topics")

    if isinstance(topics, str):
        topics = [topic.strip() for topic in topics.split(",")]

    async with Client(host, int(port), username=username, password=password) as client:
        for topic in topics:
            async with client.filtered_messages(topic) as messages:
                _logger.debug("Subscribing to topic: %s", topic)
                await client.subscribe(topic)
                async for message in messages:  # pragma: no cover
                    yield {
                        "timestamp": datetime.now(timezone.utc),
                        "name": f"{prefix}.{message.topic}",
                        "value": message.payload.decode(),
                    }



import asyncio
import logging
from typing import Dict
from typing import List
from typing import Set

from senor_octopus.graph import Source

_logger = logging.getLogger(__name__)


SLEEP_TIME = 5


class Scheduler:
    def __init__(self, dag: Set[Source]):
        self.dag = dag
        self.tasks: List[asyncio.Task] = []
        self.cancelled = False

    async def run(self) -> None:
        if not self.dag:
            _logger.info("Nothing to run")
            return

        _logger.info("Starting scheduler")
        loop = asyncio.get_event_loop()

        event_nodes = {node for node in self.dag if not node.schedule}
        for node in event_nodes:
            _logger.debug("Starting %s", node.name)
            task = asyncio.create_task(node.run())
            self.tasks.append(task)

        schedule_nodes = {node for node in self.dag if node.schedule}
        schedules: Dict[str, float] = {}
        while not self.cancelled:
            for node in schedule_nodes:
                now = loop.time()
                delay = node.schedule.next(default_utc=False)
                when = now + delay

                if node.name not in schedules:
                    _logger.info(f"Scheduling {node.name} to run in {delay} seconds")
                    schedules[node.name] = when
                elif schedules[node.name] <= now:
                    _logger.info(f"Running {node.name}")
                    task = asyncio.create_task(node.run())
                    self.tasks.append(task)
                    del schedules[node.name]

            _logger.debug(f"Sleeping for {SLEEP_TIME} seconds")
            await asyncio.sleep(SLEEP_TIME)

    def cancel(self) -> None:
        for task in self.tasks:
            task.cancel()
        self.cancelled = True



import logging
from typing import Optional

from asyncio_mqtt import Client
from senor_octopus.types import Stream

_logger = logging.getLogger(__name__)


async def mqtt(
    stream: Stream,
    topic: str,
    host: str = "localhost",
    port: str = "1883",
    username: Optional[str] = None,
    password: Optional[str] = None,
    qos: str = "1",
) -> None:
    async with Client(host, int(port), username=username, password=password) as client:
        async for event in stream:  # pragma: no cover
            await client.publish(topic, event["value"], qos=int(qos))



from dataclasses import dataclass
from datetime import datetime
from datetime import timezone
from typing import List
from typing import Union

import pytest
from asyncstdlib.builtins import aiter
from freezegun import freeze_time
from senor_octopus.sources.mqtt import mqtt


@dataclass
class FakeMessage:
    topic: str
    payload: bytes


messages = [FakeMessage("topic/1", str(i).encode("utf-8")) for i in range(3)]


@freeze_time("2021-01-01")
@pytest.mark.asyncio
async def test_mqtt(mocker) -> None:
    mock = mocker.patch("senor_octopus.sources.mqtt.Client")
    mock.return_value.__aenter__.return_value = mocker.MagicMock()
    mock.return_value.__aenter__.return_value.subscribe = mocker.AsyncMock()
    mock.return_value.__aenter__.return_value.filtered_messages.return_value.__aenter__ = (
        mocker.AsyncMock()
    )
    topics: Union[str, List[str]]

    mock.return_value.__aenter__.return_value.filtered_messages.return_value.__aenter__.return_value = aiter(
        messages,
    )
    topics = "topic/#"
    events = [event async for event in mqtt(topics)]
    assert events == [
        {
            "timestamp": datetime(2021, 1, 1, 0, 0, tzinfo=timezone.utc),
            "name": "hub.mqtt.topic/1",
            "value": "0",
        },
        {
            "timestamp": datetime(2021, 1, 1, 0, 0, tzinfo=timezone.utc),
            "name": "hub.mqtt.topic/1",
            "value": "1",
        },
        {
            "timestamp": datetime(2021, 1, 1, 0, 0, tzinfo=timezone.utc),
            "name": "hub.mqtt.topic/1",
            "value": "2",
        },
    ]

    mock.return_value.__aenter__.return_value.filtered_messages.return_value.__aenter__.return_value = aiter(
        messages,
    )
    topics = ["topic/#"]
    events = [event async for event in mqtt(topics)]
    assert events == [
        {
            "timestamp": datetime(2021, 1, 1, 0, 0, tzinfo=timezone.utc),
            "name": "hub.mqtt.topic/1",
            "value": "0",
        },
        {
            "timestamp": datetime(2021, 1, 1, 0, 0, tzinfo=timezone.utc),
            "name": "hub.mqtt.topic/1",
            "value": "1",
        },
        {
            "timestamp": datetime(2021, 1, 1, 0, 0, tzinfo=timezone.utc),
            "name": "hub.mqtt.topic/1",
            "value": "2",
        },
    ]



import logging
from datetime import datetime
from datetime import timezone
from typing import List
from typing import Union

import cryptocompare
from senor_octopus.types import Stream

_logger = logging.getLogger(__name__)


async def crypto(
    coins: Union[str, List[str]],
    currency: str = "USD",
    prefix: str = "hub.crypto",
) -> Stream:
    _logger.info("Fetching crypto data")

    if isinstance(coins, str):
        coins = [coin.strip() for coin in coins.split(",")]

    for coin in coins:
        info = cryptocompare.get_price(coin, currency=currency, full=False)
        value = info[coin][currency]
        _logger.debug("%s: %s %s", coin, currency, value)
        yield {
            "timestamp": datetime.now(timezone.utc),
            "name": f"{prefix}.{coin}.{currency}",
            "value": value,
        }



import logging
from datetime import datetime
from datetime import timezone
from typing import List
from typing import Union

import stockquotes
from senor_octopus.types import Stream

_logger = logging.getLogger(__name__)


async def stock(symbols: Union[str, List[str]], prefix: str = "hub.stock") -> Stream:
    _logger.info("Fetching stock data")

    if isinstance(symbols, str):
        symbols = [symbol.strip() for symbol in symbols.split(",")]

    for symbol in symbols:
        ticker = stockquotes.Stock(symbol)
        for attribute in {"current_price", "increase_percent"}:
            value = getattr(ticker, attribute)
            _logger.debug(
                "%s (%s) %s: %s",
                ticker.symbol,
                ticker.name,
                attribute,
                value,
            )
            yield {
                "timestamp": datetime.now(timezone.utc),
                "name": f"{prefix}.{symbol}.{attribute}",
                "value": value,
            }
